#include "ras_phys_file.h"


bool	Ras_phys_file::startup()
{
	tiling.initialize();
	
	return true;
}


dtype	Ras_phys_file::rgeo(Coord xy)
{
	// translate the given geo-coordinate to a cell-coordinate
	return rcell(region().coords_to_point(xy));
}


void	Ras_phys_file::wgeo(Coord xy, dtype val)
{
	wcell(region().coords_to_point(xy), val);
}


dtype	Ras_phys_file::rcell(ICoord cell_xy)
{
	Ras_phtile *pht = locate_tile(cell_xy);
	
	if (pht == NULL) {
		// cell_xy is out of the region
		//std::cout << "Bad icoord: " << cell_xy << "\n";
		//assert(1==3);
		return iface.metad.null_value;
	}
	
	if (!rw_access_tile(pht)) {
		// some problem in accessing the tile
		//!!! correct solution
		assert(1==4);
		return iface.metad.null_value;
	}
	
	//std::cout << cell_xy << "\n";
	
	return pht->rlogic(pht->icoord_global_2_local(cell_xy));
}


void	Ras_phys_file::wcell(ICoord cell_xy, dtype val)
{
	Ras_phtile *pht = locate_tile(cell_xy);
	
	if (pht == NULL || !rw_access_tile(pht)) {
		// ignore
		//!!! error
		assert(1==2);
		return ;
	}
	
	pht->wlogic(pht->icoord_global_2_local(cell_xy), val);
}
	
	
Ras_phtile *Ras_phys_file::locate_tile(ICoord coord)
{
	// Input coordinate is not in expected range
	// => there can be no Phtile for such a cell
	if (!region().i_inside(coord))
		return NULL;
	
	// translate a given logic coordinate (cell xy)
	// to a particular tile_xy identification
	// operation must succeed
	ICoord tile_xy = tiling.ltranslate(coord);
	
	// ask for a phtile-object for a given tile_xy ID
	return locate_tile_xy(tile_xy);
}


Ras_phtile *Ras_phys_file::locate_tile_xy(ICoord tile_xy)
{
	// is the tile (indexed by tile_xy) registered already?
	if (!known_tiles.exists(tile_xy)) {
		// if not, then:
		// create its object with a computed geo-region
		Ras_phtile *pht = new Ras_phtile(tiling.region_of_a_tile_xy(tile_xy));
		
		// give it its ID
		pht->tile_xy = tile_xy;
		
		// offset in cells meaning
		pht->ilb = tiling.ilb_of_a_tile_xy(tile_xy);
		
		// register
		known_tiles.insert(pht);
		
		// result is ...
		return pht;
	}
	
	// already registered, take it from the map (surely exists)
	return known_tiles.get(tile_xy, true);
}


bool	Ras_phys_file::make_tile_allocated(Ras_phtile *pht)
{
	// assume: 1) pht is not allocated already
	// 2) there is no record in allocated_tiles of pht
	assert(!pht->allocated() && !allocated_tiles.exists(pht->tile_xy));
	
	// checking the max_available_tiles
	if (needs_reallocate()) {
		// need for a free tile capacity
		if (decide_and_deallocate_tiles(1) == 0) {
			// but no success...
			return false;
		}
	}
	
	// regarding the user-defaults
	if (USERDEFS.use_phtile_buffers_pool())
		// allocate the buffer using the pool
		pht->allocate_with_a_given_buffer(
			phtile_buf_pool.allocate(pht->region.countXY()));
	else
		// the tile allocates its own buffer
		pht->do_allocate();
	
	// register the allocation
	allocated_tiles.insert(pht);
	
	return true;
}


bool	Ras_phys_file::make_tile_deallocated(Ras_phtile *pht)
{
	// assume: pht is currently allocated
	assert(pht->allocated());
	
	// swap the tile if needed
	bool swap_res = iface.swap_tile(pht);
	
	// regarding the options, do:
	//!!! strictly assume that this option did not change
	//!!! during the file's operation
	if (USERDEFS.use_phtile_buffers_pool()) 
		phtile_buf_pool.put_back(
			pht->deallocate_with_a_given_buffer());
	else
		pht->do_deallocate();
	
	// deregister the allocation
	allocated_tiles.erase(pht);
	
	// resulting value: result of swap op.
	return swap_res;
}


bool	Ras_phys_file::rw_access_tile(Ras_phtile *pht)
{
	// tile is allocated, then ok...
	if (pht->allocated())
		return true;
	
	// tile is not allocated
	if (!make_tile_allocated(pht)) {
		// and failed in the allocation
		assert(1==5);
		return false;
	}
	
	// STA
	int sta_resp = iface.swapped_tile_available(pht);
	
	// any technical failure in STA op?
	if (sta_resp == RAS_IFACE_TECH_FAILED) {
		printf("Tech failed\n");
		return false;
	}
	
	// is the tile's contents available in storage
	if (sta_resp == RAS_IFACE_OK) {
		// yes, do load its contents
		if (iface.load_swapped_tile(pht) == RAS_IFACE_OK) {
			// and the tile is "loaded"
			pht->status = PHTS_LOADED;
			return true;
		} else	{
			// failed in load (indicates some problem
			// in the storage) 
			assert(1==6);
			return false;
		}
	} else {
		// the tile is NOT yet saved in the storage
		// create it a new empty contents
		pht->fill_with_null_data(iface.metad.null_value);
		
		// the tile is now "ready"
		pht->status = PHTS_READY;
		return true;
	}
    
    return false;
}


bool	Ras_phys_file::deregister_tile(Ras_phtile *pht)
{
	bool make_deal_res = true;
	
	// strictly assume, that:
	assert(known_tiles.exists(pht->tile_xy));
	assert(pht->priority==0);
	
	// is allocated, make it deallocated first
	if (pht->allocated()) 
		make_deal_res = make_tile_deallocated(pht);
	
	// deregister and remove the object
	known_tiles.erase(pht);	delete pht;
	
	return true;
}


int	Ras_phys_file::deallocate_all_allocated_tiles()
{
	Phtile_map_i iter;
	int	num_of_errors = 0;
	
	printf("Closing all allocated tiles .... %d\n", allocated_tiles.size());
	
	// the operation is done as a complete DB transaction
	iface.transaction(true);
	while (allocated_tiles.size()>0) {
		// take the tile object
		Ras_phtile *pht = allocated_tiles.begin()->second;
		
		// make the tile deallocated
		// "pht" is still registered as known_tiles !!!
		if (make_tile_deallocated(pht) == false)
			num_of_errors++;
	}
	iface.transaction(false);
	
	if (num_of_errors>0) 
		printf("!!!! %d errors\n", num_of_errors);
	
	return num_of_errors;
}


int	Ras_phys_file::close_all_tiles()
{
	printf("Closing the file ....\n");
	
	int	num_of_errors = 0;
	// at first, do deallocate all
	num_of_errors = deallocate_all_allocated_tiles();
	
	// at second, delete all phtile objects 
	
	Phtile_map_i iter;
	while (known_tiles.size()>0) {
		// take the tile object
		Ras_phtile *pht = known_tiles.begin()->second;
		if (deregister_tile(pht) == false)
			num_of_errors++;
	}
	
	if (num_of_errors>0) 
		printf("!!!! %d errors in total\n", num_of_errors);
	
	return num_of_errors;
}


void	Ras_phys_file::create_whole_raster()
{
	ICoord tl(0,0);
	ICoord tlsize = tiling.number_of_tiles();
	
	iface.transaction(true);
	
	printf("Size in tiles %d %d\n", tlsize.x, tlsize.y);
	
	for (tl.y=0; tl.y<tlsize.y; tl.y++) 
	for (tl.x=0; tl.x<tlsize.x; tl.x++) {
		//printf("In %d %d\n", tl.x, tl.y);
		Ras_phtile *pht = locate_tile_xy(tl);
		//printf("Located %d %d\n", pht->tile_xy.x, pht->tile_xy.y);
		assert(rw_access_tile(pht));
		pht->dirty = true;
		deregister_tile(pht);
	}
	
	iface.transaction(false);
}

